@@ -1,7 +1,9 @@
 """Remote and distributed numpy arrays
 """
 
-from .distob import proxy_methods, Remote, Error, scatter, gather
+from __future__ import absolute_import
+from .distob import (proxy_methods, Remote, Ref, Error, 
+                     scatter, gather, vectorize)
 import numpy as np
 from collections import Sequence
 import numbers
@@ -16,7 +18,16 @@ _TupleType = type(())
 _NewaxisType = type(np.newaxis)
 
 
-@proxy_methods(np.ndarray, include_underscore=(
+def _brief_warning(msg, stacklevel=None):
+    old_format = warnings.formatwarning
+    def brief_format(message, category, filename, lineno, line=None):
+        return '%s [%s:%d]\n' % (msg, filename, lineno)
+    warnings.formatwarning = brief_format
+    warnings.warn(msg, RuntimeWarning, stacklevel + 1)
+    warnings.formatwarning = old_format
+
+
+@proxy_methods(np.ndarray, exclude=('dtype',), include_underscore=(
     '__getitem__', '__setitem__', '__getslice__', '__setslice__'))
 class RemoteArray(Remote, object):
     """Local object representing a remote ndarray"""
@@ -33,11 +44,15 @@ class RemoteArray(Remote, object):
         # implement the python array interface
         self._array_intf = {'descr': descr, 'shape': shape, 'strides': strides,
                             'typestr': typestr, 'version': 3}
+        self.dtype = np.dtype(typestr)
 
     #If a local consumer wants direct data access via the python 
     #array interface, then ensure a local copy of the data is in memory
     def __get_array_intf(self):
-        #print('__array_interface__ requested.')
+        if not self._obcache_current:
+            msg = (u"Note: local numpy function requested local access to " +
+                    "data. Fetching data..")
+            _brief_warning(msg, stacklevel=4)
         self._fetch()
         self._array_intf['data'] = self._obcache.__array_interface__['data']
         return self._array_intf
@@ -113,6 +128,92 @@ class RemoteArray(Remote, object):
         print('In __array_wrap__')
         return np.ndarray.__array_wrap__(out_arr, context)
 
+    def expand_dims(self, axis):
+        """Insert a new axis, at a given position in the array shape
+        Args:
+          axis (int): Position (amongst axes) where new axis is to be inserted.
+        """
+        ix = [slice(None)] * self.ndim
+        ix.insert(axis, np.newaxis)
+        ix = tuple(ix)
+        return self[ix]
+
+    def concatenate(self, tup, axis=0):
+        """Join a sequence of arrays to this one.
+
+        Will aim to join `ndarray`, `RemoteArray`, and `DistArray` without
+        moving their data if they are on different engines.
+
+        Args: 
+          tup (sequence of array_like): Arrays to be joined with this one.
+            They must have the same shape as this array, except in the
+            dimension corresponding to `axis`.
+          axis (int, optional): The axis along which arrays will be joined.
+
+        Returns:
+          res `RemoteArray`, if arrays were all on the same engine
+              `DistArray`, if inputs were on different engines
+        """
+        if not isinstance(tup, Sequence):
+            tup = (tup,)
+        if tup is (None,) or len(tup) is 0:
+            return self
+        tup = (self,) + tuple(tup)
+        # convert all arguments to RemoteArrays if they are not already:
+        arrays = []
+        for ar in tup:
+            if isinstance(ar, DistArray):
+                if axis == ar._distaxis:
+                    ix_new_concat_axis = tuple(
+                        ([slice(None)]*(ar.ndim - 1)).insert(axis, np.newaxis))
+                    arrays.extend(
+                        [ra[ix_new_concat_axis] for ra in ar._subarrays])
+                else:
+                    # Since not yet implemented arrays distributed on more than
+                    # one axis, will fetch and re-scatter on the new axis.
+                    ar = gather(ar)
+                    arrays.extend(np.split(ar, ar.shape[axis], axis))
+            elif isinstance(ar, RemoteArray):
+                arrays.append(ar)
+            elif isinstance(ar, Remote):
+                arrays.append(_remote_to_array(ar))
+            elif (not isinstance(ar, Remote) and 
+                    not isinstance(ar, DistArray) and
+                    not hasattr(type(ar), '__array_interface__')):
+                arrays.append(np.array(ar))
+        rarrays = scatter(arrays) # TODO: this will only work on the client
+        # At this point rarrays is a list of RemoteArray to be concatenated
+        refs = [ra._ref for ra in rarrays]
+        eng_id = refs[0].engine_id
+        if all(ref.engine_id == eng_id for ref in refs):
+            # Arrays to be joined are all on the same engine
+            from distob import engine
+            if eng_id == engine.id:
+                # All are on local host. Concatenate the local objects:
+                return concatenate([engine[r.object_id] for r in refs], axis)
+            else:
+                # All are on the same remote host. Concatenate to a RemoteArray
+                ids = [ref.object_id for ref in refs]
+                dv = engine._client[eng_id]
+                def remote_concat(ids, axis):
+                    from distob import engine
+                    ar = np.concatenate([engine[id] for id in ids], axis)
+                    return Ref(ar)
+                ref = dv.apply_sync(remote_concat, ids, axis)
+                return RemoteArray(ref)
+        else:
+            # Arrays to be joined are on different engines.
+            concatenation_axis_lengths = [ra.shape[axis] for ra in rarrays]
+            if not all(n == 1 for n in concatenation_axis_lengths):
+                #TODO: should do it remotely, using distob.split()
+                msg = (u'Concatenating remote axes with lengths other than 1.'
+                        'current implementation will fetch all data locally!')
+                warnings.warn(msg, RuntimeWarning)
+                array = concatenate(gather(rarrays), axis)
+                return scatter(array, axis)
+            else:
+                return DistArray(rarrays, axis)
+
 
 class DistArray(object):
     """Local object representing a single array distributed on multiple engines
@@ -123,10 +224,10 @@ class DistArray(object):
         """Make a DistArray from a list of existing remote numpy.ndarrays
 
         Args:
-          subarrays (list of Ref): list of references to subarrays (possibly
-            remote) which form the whole array when concatenated. The subarrays
-            must all have the same shape and dtype.
-          axis (int, optional): Position of concatenation axis. 
+          subarrays (list of RemoteArray, or list of Ref to ndarray): 
+            the subarrays (possibly remote) which form the whole array when
+            concatenated. The subarrays must all have the same shape and dtype.
+          axis (int, optional): Position of the distributed axis.
             If n subarrays are given, each of shape (i1, i2, ..., iK), 
             and `axis` is not given, the resulting DistArray will have shape 
             (i1, i2, ..., iK, n). But if `axis` is given then the subarrays 
@@ -137,10 +238,13 @@ class DistArray(object):
             raise ValueError('must provide more than one subarray')
         self._obcache = None
         self._obcache_current = False
-        first_array_metadata = subarrays[0].metadata
-        if not all(ref.metadata == first_array_metadata for ref in subarrays):
+        # In the subarrays list, accept either RemoteArray or Ref to ndarray:
+        subarrays = [ra if isinstance(ra, RemoteArray)
+                     else RemoteArray(ra) for ra in subarrays]
+        first_ra_metadata = subarrays[0]._ref.metadata
+        if not all(ra._ref.metadata == first_ra_metadata for ra in subarrays):
             raise ValueError('subarrays must have same shape, strides & dtype')
-        descr, subshape, substrides, typestr = first_array_metadata
+        descr, subshape, substrides, typestr = first_ra_metadata
         itemsize = int(typestr[2:])
         if axis is None:
             self._distaxis = len(subshape)
@@ -157,7 +261,8 @@ class DistArray(object):
         strides = list(substrides)
         strides.insert(self._distaxis, int(np.product(subshape)*itemsize))
         strides = tuple(strides)
-        self._subarrays = [RemoteArray(ref) for ref in subarrays]
+        self.dtype = np.dtype(typestr)
+        self._subarrays = subarrays
         # a surrogate ndarray to help with slicing of the distributed axis:
         self._placeholders = np.array(range(len(subarrays)), dtype=int)
         # implement the python array interface
@@ -168,8 +273,8 @@ class DistArray(object):
         """update local cached copy of the real object"""
         if not self._obcache_current:
             ax = self._distaxis
-            self._obcache = np.concatenate(
-                    [np.expand_dims(ra._ob, ax) for ra in self._subarrays], ax)
+            self._obcache = concatenate(
+                    [expand_dims(ra._ob, ax) for ra in self._subarrays], ax)
             self._obcache_current = True
 
     def __ob(self):
@@ -195,7 +300,13 @@ class DistArray(object):
 
     def __repr__(self):
         classname = self.__class__.__name__
-        ix = tuple(np.meshgrid(*(((0, 1, -2, -1),)*self.ndim), indexing='ij'))
+        selectors = []
+        for i in range(self.ndim):
+            if self.shape[i] > 3:
+                selectors.append((0, 1, -2, -1))
+            else:
+                selectors.append(tuple(range(self.shape[i])))
+        ix = tuple(np.meshgrid(*selectors, indexing='ij'))
         corners = gather(self[ix])
         def _repr_nd(corners, shape, indent):
             """Recursively generate abbreviated text representation of array"""
@@ -321,14 +432,18 @@ class DistArray(object):
             otherix = index[0:distaxis] + index[(distaxis+1):]
             ixlist = self._placeholders[distix]
             if isinstance(ixlist, numbers.Number):
-                ixlist = [ixlist]
+                # distributed axis has been sliced away: return a RemoteArray
+                return self._subarrays[ixlist][otherix]
             some_subarrays = [self._subarrays[i] for i in ixlist]
             result_ras = [ra[otherix] for ra in some_subarrays]
-            new_distaxis = distaxis
             if len(result_ras) is 1:
-                return result_ras[0] #no longer distributed: return RemoteArray
+                # no longer distributed: return a RemoteArray
+                return expand_dims(result_ras[0], distaxis)
             else:
-                return DistArray([ra._ref for ra in result_ras], new_distaxis)
+                axes_removed = sum(1 for x in index[:distaxis] if isinstance(
+                        x, numbers.Integral))
+                new_distaxis = distaxis - axes_removed
+                return DistArray(result_ras, new_distaxis)
         else:
             # advanced integer slicing
             is_fancy = tuple(not isinstance(x, _SliceType) for x in index)
@@ -361,7 +476,7 @@ class DistArray(object):
                 else:
                     earlier_fancy = len([i for i in fancy_pos if i < distaxis])
                     new_distaxis = distaxis - earlier_fancy + idim
-                remove_axis = ((slice(None),)*(new_distaxis - 1) + (0,) + 
+                remove_axis = ((slice(None),)*(new_distaxis) + (0,) + 
                                (slice(None),)*(self.ndim - new_distaxis - 1))
                 result_ras = [ra[subix][remove_axis] for ra in some_subarrays]
             else:
@@ -369,10 +484,11 @@ class DistArray(object):
                 nonconstant_ix_axes = []
                 for j in range(idim):
                     n = distix.shape[j]
-                    partix = np.split(distix, n, axis=j)
-                    if not all(np.array_equal(
-                            partix[0], partix[i]) for i in range(1, n)):
-                        nonconstant_ix_axes.append(j)
+                    if n > 0:
+                        partix = np.split(distix, n, axis=j)
+                        if not all(np.array_equal(
+                                partix[0], partix[i]) for i in range(1, n)):
+                            nonconstant_ix_axes.append(j)
                 if len(nonconstant_ix_axes) <= 1:
                     # then we can apply the indexing without moving data
                     if len(nonconstant_ix_axes) is 0:
@@ -434,7 +550,7 @@ class DistArray(object):
                 # no longer distributed: return a RemoteArray
                 return expand_dims(result_ras[0], new_distaxis)
             else:
-                return DistArray([ra._ref for ra in result_ras], new_distaxis)
+                return DistArray(result_ras, new_distaxis)
 
     def __setitem__(self, index, value):
         """Assign to the sliced item"""
@@ -460,6 +576,148 @@ class DistArray(object):
 
     T = property(fget=transpose)
 
+    @classmethod
+    def __distob_vectorize__(cls, f):
+        """Upgrades a normal function f to act on a DistArray in parallel
+
+        Args:
+          f (callable): ordinary function which expects as its first 
+            argument an array (of the same shape as our subarrays)
+
+        Returns:
+          vf (callable): new function that takes a DistArray as its first
+            argument. ``vf(distarray)`` will do the computation ``f(subarray)``
+            on each subarray in parallel and if possible will return another 
+            DistArray. (otherwise will return a list with the result for each 
+            subarray).
+        """
+        def vf(self, *args, **kwargs):
+            refs = [ra._ref for ra in self._subarrays]
+            from distob import engine
+            dv = engine._client[:]
+            def remote_f(object_id, *args, **kwargs):
+                import distob
+                result = f(distob.engine[object_id], *args, **kwargs)
+                if type(result) in distob.engine.proxy_types:
+                    return Ref(result)
+                else:
+                    return result
+            results = []
+            for ref in refs:
+                dv.targets = ref.engine_id
+                ar = dv.apply_async(remote_f, ref.object_id, *args, **kwargs)
+                results.append(ar)
+            for i in range(len(results)):
+                ar = results[i]
+                ar.wait()
+                results[i] = ar.r
+                if isinstance(results[i], Ref):
+                    ref = results[i]
+                    RemoteClass = engine.proxy_types[ref.type]
+                    results[i] = RemoteClass(ref)
+            if (all(isinstance(r, RemoteArray) for r in results) and
+                    all(r.shape == results[0].shape for r in results)):
+                # Then we can join the results and return a DistArray.
+                # We will keep the same axis distributed as in the input,
+                # unless the results have more dimensions than the input.
+                res_subshape = results[0].shape
+                if len(res_subshape) > self.ndim - 1:
+                    res_distaxis = len(res_subshape)
+                else:
+                    res_distaxis = self._distaxis
+                newaxes = res_distaxis - len(res_subshape)
+                if newaxes >= 1:
+                    ix = ((slice(None),) * len(res_subshape) + 
+                          (np.newaxis,) * newaxes)
+                    results = [r[ix] for r in results]
+                return DistArray(results, res_distaxis)
+            elif all(isinstance(r, numbers.Number) for r in results):
+                return np.array(results)
+            else:
+                return results  # list
+        if hasattr(f, '__name__'):
+            vf.__name__ = 'v' + f.__name__
+            f_str = f.__name__ + '()'
+        else:
+            f_str = 'callable'
+        doc = u"""Apply %s in parallel to a DistArray\n
+               Args:
+                 da (DistArray)
+                 other args are the same as for %s
+               """ % (f_str, f_str)
+        if hasattr(f, '__doc__') and f.__doc__ is not None:
+            doc = doc.rstrip() + (' detailed below:\n----------\n' + f.__doc__)
+        vf.__doc__ = doc
+        return vf
+
+    def __distob_gather__(self):
+        return self._ob
+
+    def expand_dims(self, axis):
+        """Insert a new axis, at a given position in the array shape
+        Args:
+          axis (int): Position (amongst axes) where new axis is to be inserted.
+        """
+        if axis <= self._distaxis:
+            subaxis = axis
+            new_distaxis = self._distaxis + 1
+        else:
+            subaxis = axis - 1
+            new_distaxis = self._distaxis
+        new_subarrays = [expand_dims(ra, subaxis) for ra in self._subarrays]
+        return DistArray(new_subarrays, new_distaxis)
+
+    def concatenate(self, tup, axis=0):
+        """Join a sequence of arrays to this one.
+
+        Will aim to join `ndarray`, `RemoteArray`, and `DistArray` without
+        moving their data if they are on different engines.
+
+        Args: 
+          tup (sequence of array_like): Arrays to be joined with this one.
+            They must have the same shape as this array, except in the
+            dimension corresponding to `axis`.
+          axis (int, optional): The axis along which arrays will be joined.
+
+        Returns:
+          res `RemoteArray`, if arrays were all on the same engine
+              `DistArray`, if inputs were on different engines
+        """
+        if not isinstance(tup, Sequence):
+            tup = (tup,)
+        if tup is (None,) or len(tup) is 0:
+            return self
+        if axis == self._distaxis:
+            ix_new_concat_axis = tuple(
+                    ([slice(None)]*(self.ndim - 1)).insert(axis, np.newaxis))
+            split_self = [ra[ix_new_concat_axis] for ra in ar._subarrays]
+        else:
+            # Since we have not yet implemented arrays distributed on more than
+            # one axis, will fetch subarrays and re-scatter on the new axis.
+            split_self = split(self._ob, self.shape[axis], axis)
+        first = split_self[0]
+        others = tuple(split_self[1:]) + tup
+        first = scatter(first)
+        return first.concatenate(others, axis)
+
+    def mean(self, axis=None, dtype=None, out=None):
+        """Compute the arithmetic mean along the specified axis."""
+        if axis == self._distaxis:
+            return np.mean(self._ob, axis, dtype, out)
+        elif axis is not None:
+            if out is not None:
+                out[:] = vectorize(np.mean)(self, axis, dtype)
+                return out
+            else:
+                return vectorize(np.mean)(self, axis, dtype)
+        else:
+            if out is not None:
+                out[:] = vectorize(np.mean)(self, axis, dtype).mean()
+                return out
+            else:
+                return vectorize(np.mean)(self, axis, dtype).mean()
+
+
 
 def transpose(a, axes=None):
     """Returns a view of the array with axes transposed.
@@ -532,22 +790,10 @@ def expand_dims(a, axis):
       a (array_like): Input array.
       axis (int): Position (amongst axes) where new axis is to be inserted.
     """
+    if hasattr(a, 'expand_dims') and hasattr(type(a), '__array_interface__'):
+        return a.expand_dims(axis)
     if isinstance(a, np.ndarray):
         return np.expand_dims(a, axis)
-    if isinstance(a, RemoteArray):
-        ix = [slice(None)] * a.ndim
-        ix.insert(axis, np.newaxis)
-        ix = tuple(ix)
-        return a[ix]
-    if isinstance(a, DistArray):
-        if axis <= a._distaxis:
-            subaxis = axis
-            new_distaxis = a._distaxis + 1
-        else:
-            subaxis = axis - 1
-            new_distaxis = a._distaxis
-        new_subarrays = [expand_dims(ra, subaxis) for ra in a._subarrays]
-        return DistArray([ra._ref for ra in new_subarrays], new_distaxis)
 
 
 def _remote_to_array(remote):
@@ -555,10 +801,11 @@ def _remote_to_array(remote):
     if isinstance(remote, RemoteArray):
         return remote
     else:
-        dv = distob.engine._client[remote._ref.engine_id]
+        from distob import engine
+        dv = engine._client[remote._ref.engine_id]
         def remote_array(object_id):
             return Ref(np.array(engine[object_id]))
-        return dv.apply_sync(remote_array, remote._ref.object_id)
+        return RemoteArray(dv.apply_sync(remote_array, remote._ref.object_id))
 
 
 def concatenate(tup, axis=0):
@@ -576,52 +823,21 @@ def concatenate(tup, axis=0):
            `RemoteArray`, if inputs were all on the same remote engine
            `DistArray`, if inputs were already scattered on different engines
     """
+    if len(tup) is 0:
+        raise ValueError('need at least one array to concatenate')
+    first = tup[0]
+    others = tup[1:]
+    if (hasattr(first, 'concatenate') and 
+            hasattr(type(first), '__array_interface__')):
+        # This case includes RemoteArray and DistArray
+        return first.concatenate(others, axis)
     if all(isinstance(ar, np.ndarray) for ar in tup):
         return np.concatenate(tup, axis)
-    # otherwise convert all arguments to RemoteArrays if they are not already:
-    arrays = []
-    for ar in tup:
-        if isinstance(ar, DistArray):
-            if axis == ar._distaxis:
-                ix_new_concat_axis = tuple(
-                        ([slice(None)]*(ar.ndim - 1)).insert(axis, np.newaxis))
-                arrays.extend([ra[ix_new_concat_axis] for ra in ar._subarrays])
-            else:
-                # Since not yet implemented arrays distributed on more than
-                # one axis, will have to fetch and re-scatter on the new axis.
-                ar = gather(ar)
-                arrays.extend(np.split(ar, ar.shape[axis], axis))
-        elif isinstance(ar, Remote) and not isinstance(ar, RemoteArray):
-            arrays += _remote_to_array(ar)
-        elif (not isinstance(ar, Remote) and 
-                not isinstance(ar, DistArray) and
-                not hasattr(type(ar), '__array_interface__')):
-            arrays += np.array(ar)
-    rarrays = scatter(arrays)
-    # At this point rarrays should be a list of RemoteArray to be concatenated
-    concatenation_axis_lengths = [ra.shape[axis] for ra in rarrays]
-    if not all(n == 1 for n in concatenation_axis_lengths):
-        message = (u'So far have only implemented support for joining arrays '
-                    'where the axes being concatenated have length 1.')
-        raise Error(message)
-    refs = [ra._ref for ra in rarrays]
-    eng_id = refs[0].engine_id
-    if all(ref.engine_id == eng_id for ref in refs):
-        # Arrays to be joined are all on the same engine
-        if eng_id == engine.id:
-            # All are on local host. Concatenate to a ndarray.
-            return np.concatenate([engine[r.object_id] for r in refs], axis)
-        else:
-            # All are on the same remote host. Concatenate to a RemoteArray.
-            ids = [ref.object_id for ref in refs]
-            dv = distob.engine._client[eng_id]
-            def remote_concat(ids, axis):
-                ar = np.concatenate([engine[id] for id in ids], axis)
-                return Ref(ar)
-            return dv.apply_sync(remote_concat, ids, axis)
+    if isinstance(first, Remote):
+        first = _remote_to_array(first)
     else:
-        # Arrays to be joined are on different engines.
-        return DistArray(refs, axis)
+        first = scatter(np.array(first)) #TODO: this only works on the client
+    return first.concatenate(others, axis)
 
 
 def vstack(tup):
@@ -683,3 +899,37 @@ def dstack(tup):
         if arrays[i].ndim is 2:
             arrays[i] = arrays[i][:, :, np.newaxis]
     return concatenate(arrays, axis=2)
+
+
+def split(a, indices_or_sections, axis=0):
+    if type(a) is np.ndarray: # deliberately excluding subclasses
+        return np.split(a, indices_or_sections, axis)
+    else:
+        if not isinstance(indices_or_sections, numbers.Integral):
+            raise Error('splitting by array of indices is not yet implemented')
+        n = indices_or_sections
+        if self.shape[axis] % n != 0:
+            raise ValueError("Array split doesn't result in an equal division")
+        step = self.shape[axis] / n
+        pieces = []
+        start = 0
+        while start < self.shape[axis]:
+            stop = start + step
+            ix = [slice(None)] * self.ndim
+            ix[axis] = slice(start, stop)
+            ix = tuple(ix)
+            pieces.append(self[ix])
+            start += step
+        return pieces
+
+
+def vsplit(a, indices_or_sections):
+    return split(a, indices_or_sections, axis=0)
+
+
+def hsplit(a, indices_or_sections):
+    return split(a, indices_or_sections, axis=1)
+
+
+def dsplit(a, indices_or_sections):
+    return split(a, indices_or_sections, axis=2)
