@@ -11,6 +11,8 @@ This module implements the core functionality of distob.
 Functions:
   scatter(obj)  Distribute an object to remote iPython engines, return a proxy.
   gather(obj)  Fetch back a distributed object, making it local again.
+  vectorize(f)  Upgrade normal function f to act in parallel on distributed
+    lists or arrays
   call_all(sequence, methodname, *args, **kwargs) Call method on each element. 
 
 Classes:
@@ -30,12 +32,10 @@ import copy
 import warnings
 import importlib
 import collections
+import numbers
 
-try:
+if distob._have_numpy:
     import numpy as np
-    _have_numpy = True
-except ImportError:
-    pass
 
 # types for compatibility across python 2 and 3
 try:
@@ -317,10 +317,16 @@ def _setup_engines(client=None):
       client (IPython.parallel.client, optional): If None, will create a client         using the default IPython profile.
     """
     if not client:
-        client = parallel.Client()
+        try:
+            client = parallel.Client()
+        except:
+            raise DistobClusterError(
+                u"""Could not connect to an IPython parallel cluster. Make
+                 sure a cluster is started (e.g. to use the CPUs of a
+                 single computer, can type 'ipcluster start')""")
     ids = client.ids
     if not ids:
-        raise DistobClusterError('No IPython compute engines are available')
+        raise DistobClusterError(u'No IPython compute engines are available')
     dv = client[ids]
     dv.use_dill()
     with dv.sync_imports(quiet=True):
@@ -525,12 +531,15 @@ class Remote(object):
             self._obcache_current = True
 
     def __ob(self):
-        """return a copy of the real object"""
+        """return a local copy of the real object"""
         self._fetch()
         return self._obcache
 
     _ob = property(fget=__ob)
 
+    def __distob_gather__(self):
+        return self._ob
+
     def __copy__(self):
         newref = copy.copy(self._ref)
         obj = self.__class__.__new__(self.__class__, newref)
@@ -670,7 +679,9 @@ def _async_scatter(obj):
     """
     #TODO Instead of special cases for strings and Remote, should have a
     #     list of types that should not be proxied, inc. strings and Remote.
-    if isinstance(obj, Remote):
+    if (isinstance(obj, Remote) or 
+            isinstance(obj, numbers.Number) or 
+            obj is None):
         return obj
     if (isinstance(obj, collections.Sequence) and 
             not isinstance(obj, string_types)):
@@ -695,20 +706,20 @@ def _async_scatter(obj):
 _async_scatter.next_engine = 0
 
 
-def _ars_to_proxies(ars, obj):
-    """wait for async results and return proxy objects to replace obj
+def _ars_to_proxies(ars):
+    """wait for async results and return proxy objects
     Args: 
       ars: AsyncResult (or sequence of AsyncResults), each result type ``Ref``.
-      obj: original object being scattered (or sequence of objects)
     Returns:
-      Remote* proxy object (or sequence of them) with same interface as obj
+      Remote* proxy object (or list of them)
     """
-    if isinstance(ars, Remote):
+    if isinstance(ars, Remote) or ars is None:
         return ars
     elif isinstance(ars, collections.Sequence):
+        res = []
         for i in range(len(ars)):
-            obj[i] = _ars_to_proxies(ars[i], obj[i])
-        return obj
+            res.append(_ars_to_proxies(ars[i]))
+        return res
     elif isinstance(ars, parallel.AsyncResult):
         ars.wait()
         ref = ars.r
@@ -741,6 +752,8 @@ def _scatter_ndarray(ar, axis=None):
                     'chunks of length 1. axis %d is long (%d), so may be slow '
                     'distributing the array along this axis.' % (axis, n))
         warnings.warn(message, RuntimeWarning)
+    if n is 1:
+        return scatter([ar])[0]
     if isinstance(ar, DistArray):
         if axis == ar._distaxis:
             return ar
@@ -756,8 +769,7 @@ def _scatter_ndarray(ar, axis=None):
         index = (s,)*axis + (i,) + (s,)*(ndim - axis - 1)
         subarrays.append(ar[index])
     subarrays = scatter(subarrays)
-    refs = [ra._ref for ra in subarrays]
-    return DistArray(refs, axis)
+    return DistArray(subarrays, axis)
 
 
 def scatter(obj, axis=None):
@@ -768,7 +780,9 @@ def scatter(obj, axis=None):
         specifying along which axis to split the array to distribute it. 
         If None, the default is to split along the last axis.
     """
-    if _have_numpy and (isinstance(obj, np.ndarray) or
+    if hasattr(obj, '__distob_scatter__'):
+        return obj.__distob_scatter__(axis)
+    if distob._have_numpy and (isinstance(obj, np.ndarray) or
                         hasattr(type(obj), '__array_interface__')):
         return _scatter_ndarray(obj, axis)
     if axis is not None:
@@ -776,26 +790,92 @@ def scatter(obj, axis=None):
     elif isinstance(obj, Remote):
         return obj
     ars = _async_scatter(obj)
-    proxy_obj = _ars_to_proxies(ars, obj)
+    proxy_obj = _ars_to_proxies(ars)
     return proxy_obj
 
 
 def gather(obj):
     """Retrieve objects that have been distributed, making them local again"""
-    if _have_numpy:
-        from .arrays import DistArray
-        if isinstance(obj, DistArray):
-            return obj._ob
-    if not isinstance(obj, Remote) and (
-            isinstance(obj, string_types) or (
-                not isinstance(obj, collections.Sequence))):
-        return obj
-    elif not isinstance(obj, Remote):
-        for i in range(len(obj)):
-            obj[i] = gather(obj[i])
-        return None
+    if hasattr(obj, '__distob_gather__'):
+        return obj.__distob_gather__()
+    elif (isinstance(obj, collections.Sequence) and 
+            not isinstance(obj, string_types)):
+        return [gather(subobj) for subobj in obj]
     else:
-        return obj._ob
+        return obj
+
+
+def vectorize(f):
+    """Upgrade normal function f to act in parallel on distibuted lists/arrays
+
+    Args:
+      f (callable): an ordinary function which expects as its first argument a
+        single object, or a numpy array of N dimensions.
+
+    Returns:
+      vf (callable): new function that takes as its first argument a list of
+        objects, or a array of N+1 dimensions. ``vf()`` will do the
+        computation ``f()`` on each part of the input in parallel and will
+        return a list of results, or a distributed array of results.
+    """
+    def vf(obj, *args, **kwargs):
+        # user classes can customize how to vectorize a function:
+        if hasattr(obj, '__distob_vectorize__'):
+            return obj.__distob_vectorize__(f)(obj, *args, **kwargs)
+        def remote_f(object_id, *args, **kwargs):
+            result = f(distob.engine[object_id], *args, **kwargs)
+            if type(result) in distob.engine.proxy_types:
+                return Ref(result)
+            else:
+                return result
+        if isinstance(obj, Remote):
+            dv = distob.engine._client[obj._ref.engine_id]
+            r = dv.apply_sync(remote_f, obj._ref.object_id, *args, **kwargs)
+            if isinstance(r, Ref):
+                RemoteClass = distob.engine.proxy_types[r.type]
+                return RemoteClass(r)
+            else:
+                return r
+        elif distob._have_numpy and (isinstance(obj, np.ndarray) or
+                 hasattr(type(obj), '__array_interface__')):
+            distarray = scatter(obj, axis=None)
+            return vf(distarray, *args, **kwargs)
+        elif isinstance(obj, collections.Sequence):
+            inputs = scatter(obj)
+            dv = distob.engine._client[:]
+            results = []
+            for ref in refs:
+                dv.targets = ref.engine_id
+                ar = dv.apply_async(remote_f, ref.object_id, *args, **kwargs)
+                results.append(ar)
+            for i in range(len(results)):
+                ar = results[i]
+                ar.wait()
+                results[i] = ar.r
+                if isinstance(results[i], Ref):
+                    ref = results[i]
+                    RemoteClass = distob.engine.proxy_types[ref.type]
+                    results[i] = RemoteClass(ref)
+            return results
+    if hasattr(f, '__name__'):
+        vf.__name__ = 'v' + f.__name__
+        f_str = f.__name__ + '()'
+    else:
+        f_str = 'callable'
+    doc = u"""Apply %s in parallel to a list or array\n
+           Args:
+             obj (Sequence of objects or an array)
+             other args are the same as for %s
+           """ % (f_str, f_str)
+    if hasattr(f, '__doc__') and f.__doc__ is not None:
+        doc = doc.rstrip() + (' detailed below:\n----------\n' + f.__doc__)
+    vf.__doc__ = doc
+    return vf
+
+
+def apply(f, obj, *args, **kwargs):
+    """Apply a function in parallel to each element of the input"""
+    return vectorize(f)(obj, *args, **kwargs)
 
 
 def call_all(sequence, method_name, *args, **kwargs):
